@@ -1,13 +1,65 @@
 // TODO migrate to eva-common when released
 use crate::Error;
+use log::{info, trace};
 use serde::{ser::SerializeSeq, Deserialize, Deserializer, Serialize, Serializer};
 use std::cmp::Ordering;
-use std::collections::HashSet;
+use std::collections::{BTreeMap, HashSet};
 use std::fmt;
 use std::hash::{Hash, Hasher};
 use std::str::FromStr;
+use std::sync::Arc;
+use tokio::sync::RwLock;
 
-const ERR_PATH_MASK_EMPTY: &str = "Empty path mask";
+static ERR_PATH_MASK_EMPTY: &str = "Empty path mask";
+
+lazy_static::lazy_static! {
+    pub static ref ACL_DB: RwLock<Db> = RwLock::new(<_>::default());
+}
+
+#[derive(Debug)]
+pub struct Db {
+    acls: BTreeMap<String, Arc<Acl>>,
+    path: String,
+}
+
+impl Default for Db {
+    fn default() -> Self {
+        Self {
+            acls: <_>::default(),
+            path: String::new(),
+        }
+    }
+}
+
+impl Db {
+    #[inline]
+    pub fn get_acl(&self, user: &str) -> Option<Arc<Acl>> {
+        self.acls.get(user).cloned()
+    }
+    #[inline]
+    pub fn has_acl(&self, user: &str) -> bool {
+        self.acls.contains_key(user)
+    }
+    #[inline]
+    pub fn set_path(&mut self, path: &str) {
+        self.path = path.to_owned();
+    }
+    /// # Errors
+    ///
+    /// Will return err on file read / deserialize error
+    pub async fn reload(&mut self) -> Result<(), Error> {
+        info!("loading ACL {}", self.path);
+        let acls: BTreeMap<String, Acl> =
+            serde_yaml::from_str(&tokio::fs::read_to_string(&self.path).await?)?;
+        self.acls.clear();
+        // TODO optimize when pop available
+        for (user, acl) in acls {
+            self.acls.insert(user, Arc::new(acl));
+        }
+        trace!("{:?}", self.acls);
+        Ok(())
+    }
+}
 
 #[derive(Debug, Deserialize)]
 #[serde(deny_unknown_fields)]
