@@ -13,6 +13,7 @@ use tokio_rustls::TlsAcceptor;
 
 use psrt::comm::SStream;
 
+#[cfg(feature = "cluster")]
 use std::collections::BTreeMap;
 use std::collections::HashSet;
 use std::net::SocketAddr;
@@ -21,8 +22,9 @@ use std::sync::atomic;
 use std::sync::Arc;
 use std::time::{Duration, Instant};
 
-use psrt::pubsub::{ServerClient, ServerClientDB, ServerClientDBStats, Token};
+use psrt::pubsub::{ServerClient, ServerClientDB, ServerClientDBStats};
 use psrt::reduce_timeout;
+use psrt::token::Token;
 
 use psrt::RESPONSE_ERR;
 use psrt::RESPONSE_ERR_ACCESS;
@@ -40,7 +42,7 @@ use psrt::OP_UNSUBSCRIBE;
 
 use psrt::{CONTROL_HEADER, DATA_HEADER};
 
-use psrt::acl::Acl;
+use psrt::acl::{self, ACL_DB};
 use psrt::COMM_INSECURE;
 use psrt::COMM_TLS;
 
@@ -55,7 +57,7 @@ use log::{error, info, trace, warn};
 use log::{Level, LevelFilter};
 use serde::{Deserialize, Serialize};
 
-const ERR_INVALID_DATA_BLOCK: &str = "Invalid data block";
+static ERR_INVALID_DATA_BLOCK: &str = "Invalid data block";
 const MAX_AUTH_FRAME_SIZE: usize = 1024;
 const DEFAULT_UDP_FRAME_SIZE: u16 = 4096;
 
@@ -63,12 +65,12 @@ static ALLOW_ANONYMOUS: atomic::AtomicBool = atomic::AtomicBool::new(false);
 static MAX_PUB_SIZE: atomic::AtomicUsize = atomic::AtomicUsize::new(0);
 static MAX_TOPIC_LENGTH: atomic::AtomicUsize = atomic::AtomicUsize::new(0);
 
-const CONFIG_FILES: &[&str] = &["/etc/psrtd/config.yml", "/usr/local/etc/psrtd/config.yml"];
+static CONFIG_FILES: &[&str] = &["/etc/psrtd/config.yml", "/usr/local/etc/psrtd/config.yml"];
 
 #[cfg(feature = "cluster")]
-const APP_NAME: &str = "PSRT Enterprise";
+static APP_NAME: &str = "PSRT Enterprise";
 #[cfg(not(feature = "cluster"))]
-const APP_NAME: &str = "PSRT";
+static APP_NAME: &str = "PSRT";
 
 use stats::Counters;
 
@@ -115,10 +117,10 @@ macro_rules! format_login {
 lazy_static! {
     static ref DB: RwLock<ServerClientDB> = RwLock::new(<_>::default());
     static ref PASSWORD_DB: RwLock<psrt::passwords::Passwords> = RwLock::new(<_>::default());
+    static ref KEY_DB: RwLock<psrt::keys::Keys> = RwLock::new(<_>::default());
     static ref STATS_COUNTERS: RwLock<Counters> = RwLock::new(Counters::new());
     static ref PID_FILE: Mutex<Option<String>> = Mutex::new(None);
     static ref HOST_NAME: RwLock<String> = RwLock::new("unknown".to_owned());
-    static ref ACL_DB: RwLock<AclDb> = RwLock::new(AclDb::new());
     static ref UPTIME: Instant = Instant::now();
 }
 
@@ -186,41 +188,6 @@ enum StreamType {
     Data,
 }
 
-#[derive(Debug)]
-struct AclDb {
-    acls: BTreeMap<String, Arc<Acl>>,
-    path: String,
-}
-
-impl AclDb {
-    fn new() -> Self {
-        Self {
-            acls: <_>::default(),
-            path: String::new(),
-        }
-    }
-    #[inline]
-    fn get_acl(&self, user: &str) -> Option<Arc<Acl>> {
-        self.acls.get(user).cloned()
-    }
-    #[inline]
-    fn set_path(&mut self, path: &str) {
-        self.path = path.to_owned();
-    }
-    async fn reload(&mut self) -> Result<(), Error> {
-        info!("loading ACL {}", self.path);
-        let acls: BTreeMap<String, Acl> =
-            serde_yaml::from_str(&tokio::fs::read_to_string(&self.path).await?)?;
-        self.acls.clear();
-        // TODO optimize when pop available
-        for (user, acl) in acls {
-            self.acls.insert(user, Arc::new(acl));
-        }
-        trace!("{:?}", self.acls);
-        Ok(())
-    }
-}
-
 #[allow(clippy::cast_possible_truncation)]
 async fn push_to_subscribers(
     subscribers: &HashSet<ServerClient>,
@@ -255,7 +222,7 @@ async fn process_control(
     mut stream: SStream,
     client: ServerClient,
     addr: SocketAddr,
-    acl: Arc<Acl>,
+    acl: Arc<acl::Acl>,
     timeout: Duration,
 ) -> Result<(), Error> {
     loop {
@@ -535,7 +502,7 @@ pub async fn authenticate(login: &str, password: &str) -> bool {
 }
 
 #[inline]
-pub async fn get_acl(login: &str) -> Option<Arc<Acl>> {
+pub async fn get_acl(login: &str) -> Option<Arc<acl::Acl>> {
     acl_db!().get_acl(if login.is_empty() { "_" } else { login })
 }
 
@@ -613,6 +580,7 @@ async fn launch_data_stream(
     let token = Token::from(buf);
     let (tx, rx) = async_channel::bounded(psrt::pubsub::get_data_queue_size());
     let mut buf: [u8; 1] = [0];
+    let latency_warn = psrt::pubsub::get_latency_warn();
     stream
         .read_with_timeout(&mut buf, reduce_timeout(timeout, op_start))
         .await?;
@@ -686,7 +654,7 @@ async fn launch_data_stream(
                                     );
                                 };
                                 trace!("latency: {} \u{3bc}s", latency_mks);
-                                if latency_mks > psrt::pubsub::get_latency_warn() {
+                                if latency_mks > latency_warn {
                                     warn!(
                                         "WARNING: high latency: {} \u{3bc}s topic {} ({}@{})",
                                         latency_mks,
@@ -730,6 +698,8 @@ struct ConfigCluster {
 struct ConfigAuth {
     allow_anonymous: bool,
     password_file: Option<String>,
+    key_file: Option<String>,
+    aes_nonce: Option<String>,
     acl: String,
 }
 
@@ -925,32 +895,13 @@ async fn run_server(
     }
 }
 
-/// bool in replies = true if ack required
-async fn process_udp_packet(frame: Vec<u8>) -> Result<bool, (Error, bool)> {
-    let timestamp = now_ns();
-    if frame.len() < 5 {
-        return Err((Error::invalid_data("packet too small"), false));
-    }
-    if frame[..2] != CONTROL_HEADER {
-        return Err((Error::invalid_data("invalid header"), false));
-    }
-    if u16::from_le_bytes([frame[2], frame[3]]) != psrt::PROTOCOL_VERSION {
-        return Err((Error::invalid_data("unsupported protocol version"), false));
-    }
-    if frame[4] != 0 {
-        return Err((Error::invalid_data("invalid packet format"), false));
-    }
-    let mut sp = frame[5..].splitn(4, |n| *n == 0);
-    let login = std::str::from_utf8(
-        sp.next()
-            .ok_or_else(|| (Error::invalid_data("login missing"), false))?,
-    )
-    .map_err(|e| (e.into(), false))?;
-    let password = std::str::from_utf8(
-        sp.next()
-            .ok_or_else(|| (Error::invalid_data("password missing"), false))?,
-    )
-    .map_err(|e| (e.into(), false))?;
+async fn process_udp_block(
+    login: &str,
+    password: Option<&str>,
+    block: &[u8],
+    timestamp: u64,
+) -> Result<bool, (Error, bool)> {
+    let mut sp = block.splitn(2, |n: &u8| *n == 0);
     let buf = sp
         .next()
         .ok_or_else(|| (Error::invalid_data("data block missing"), false))?;
@@ -972,12 +923,14 @@ async fn process_udp_packet(frame: Vec<u8>) -> Result<bool, (Error, bool)> {
     let data = sp
         .next()
         .ok_or_else(|| (Error::invalid_data("data missing"), need_ack))?;
-    if login.is_empty() && password.is_empty() {
-        if !ALLOW_ANONYMOUS.load(atomic::Ordering::SeqCst) {
-            return Err((Error::access("anonymous access failed"), need_ack));
+    if let Some(password) = password {
+        if login.is_empty() && password.is_empty() {
+            if !ALLOW_ANONYMOUS.load(atomic::Ordering::SeqCst) {
+                return Err((Error::access("anonymous access failed"), need_ack));
+            }
+        } else if !authenticate(login, password).await {
+            return Err((Error::access("authentication failed"), need_ack));
         }
-    } else if !authenticate(login, password).await {
-        return Err((Error::access("authentication failed"), need_ack));
     }
     let acl = if let Some(acl) = get_acl(login).await {
         acl
@@ -1008,6 +961,59 @@ async fn process_udp_packet(frame: Vec<u8>) -> Result<bool, (Error, bool)> {
     Ok(need_ack)
 }
 
+/// bool in replies = true if ack required
+async fn process_udp_packet(frame: Vec<u8>) -> Result<bool, (Error, bool)> {
+    let timestamp = now_ns();
+    if frame.len() < 5 {
+        return Err((Error::invalid_data("packet too small"), false));
+    }
+    if frame[..2] != CONTROL_HEADER {
+        return Err((Error::invalid_data("invalid header"), false));
+    }
+    if u16::from_le_bytes([frame[2], frame[3]]) != psrt::PROTOCOL_VERSION {
+        return Err((Error::invalid_data("unsupported protocol version"), false));
+    }
+    let etp = psrt::keys::EncryptionType::from_byte(frame[4]).map_err(|e| (e, false))?;
+    trace!("UDP packet encryption: {:?}", etp);
+    let mut sp = frame[5..].splitn(2, |n: &u8| *n == 0);
+    let login = std::str::from_utf8(
+        sp.next()
+            .ok_or_else(|| (Error::invalid_data("login / key id missing"), false))?,
+    )
+    .map_err(|e| (e.into(), false))?;
+    if etp.need_decrypt() {
+        let block = KEY_DB
+            .read()
+            .await
+            .auth_and_decr(
+                sp.next()
+                    .ok_or_else(|| (Error::invalid_data("encryption block missing"), false))?,
+                login,
+                etp,
+            )
+            .map_err(|e| (e, false))?;
+        process_udp_block(login, None, &block, timestamp).await
+    } else {
+        let block = sp
+            .next()
+            .ok_or_else(|| (Error::invalid_data("invalid packet format"), false))?;
+        let mut sp = block.splitn(2, |n: &u8| *n == 0);
+        let password = std::str::from_utf8(
+            sp.next()
+                .ok_or_else(|| (Error::invalid_data("password missing"), false))?,
+        )
+        .map_err(|e| (e.into(), false))?;
+        process_udp_block(
+            login,
+            Some(password),
+            sp.next()
+                .ok_or_else(|| (Error::invalid_data("data missing"), false))?,
+            timestamp,
+        )
+        .await
+    }
+}
+
 async fn terminate(allow_log: bool) {
     let pid_file = PID_FILE.lock().await;
     #[cfg(feature = "cluster")]
@@ -1028,7 +1034,7 @@ async fn terminate(allow_log: bool) {
 macro_rules! handle_term_signal {
     ($kind: expr, $allow_log: expr) => {
         tokio::spawn(async move {
-            trace!("starging handler for {:?}", $kind);
+            trace!("starting handler for {:?}", $kind);
             loop {
                 match signal($kind) {
                     Ok(mut v) => {
@@ -1201,10 +1207,10 @@ fn main() {
         .enable_all()
         .build()
         .unwrap();
-    macro_rules! reload_passwords {
+    macro_rules! reload_db {
         ($db: expr) => {
             if let Err(e) = $db.reload().await {
-                error!("Unable to load password file: {}", e);
+                error!("Unable to load config: {}", e);
             }
         };
     }
@@ -1218,13 +1224,25 @@ fn main() {
             let mut passwords = PASSWORD_DB.write().await;
             let password_file = format_path!(f.clone());
             passwords.set_password_file(&password_file);
-            reload_passwords!(passwords);
+            reload_db!(passwords);
+        }
+        if let Some(ref f) = config.auth.key_file {
+            let mut keys = KEY_DB.write().await;
+            let key_file = format_path!(f.clone());
+            keys.set_key_file(&key_file);
+            keys.set_nonce(config.auth.aes_nonce.as_ref().map(|h| {
+                hex::decode(h)
+                    .expect("unable to parse nonce")
+                    .try_into()
+                    .expect("invalid nonce size (12 bytes required)")
+            }));
+            reload_db!(keys);
         }
         handle_term_signal!(SignalKind::interrupt(), false);
         handle_term_signal!(SignalKind::terminate(), true);
         tokio::spawn(async move {
             let kind = SignalKind::hangup();
-            trace!("starging handler for {:?}", kind);
+            trace!("starting handler for {:?}", kind);
             loop {
                 match signal(kind) {
                     Ok(mut v) => {
@@ -1241,7 +1259,10 @@ fn main() {
                         error!("ACL reload failed: {}", e);
                     }
                     {
-                        reload_passwords!(PASSWORD_DB.write().await);
+                        reload_db!(PASSWORD_DB.write().await);
+                    }
+                    {
+                        reload_db!(KEY_DB.write().await);
                     }
                 }
             }
